import requests

from serviceprovider.ranger_models import UrlScheme
from serviceprovider.helper import get_default_logger


class HealthCheck(object):
    def __init__(self, url, scheme: UrlScheme, logger=None, data=None, headers=None, timeout=1.0, acceptable_errors=None):
        """
        :param url: url path where health pings can be done at
        :param scheme: GET / POST
        :param logger: custom logger if provided
        :param data: data if its a POST scheme
        :param headers: headers if required
        :param timeout: timeout for the ping call
        :param acceptable_errors: in case you wanna indicate that some non-2xx response is acceptable
        """
        self.url = url
        self.scheme = scheme
        self.logger = logger if logger is not None else get_default_logger()
        self.data = data
        self.timeout = timeout
        self.headers = headers if headers is not None else {"Content-Type": "application/json"}
        self.acceptable_errors = acceptable_errors if acceptable_errors is not None else []

    def is_healthy(self):
        try:
            if UrlScheme.GET == self.scheme:
                resp = requests.get(url=self.url, headers=self.headers, timeout=self.timeout)
                self.logger.info(f"Checking health at:{self.url}, URL returned:{resp.status_code}")
                return self._check_status(resp)
            elif UrlScheme.POST == self.scheme:
                resp = requests.post(url=self.url, headers=self.headers, timeout=self.timeout, data=self.data)
                self.logger.info(f"Checking health, URL returned:{resp.status_code}")
                return self._check_status(resp)
            else:
                self.logger.info(f"Invalid scheme: {self.scheme}")
        except requests.exceptions.ConnectionError:
            self.logger.warning(f"Unable to connect to healthcheck url {self.url}")
        except requests.exceptions.ReadTimeout:
            self.logger.warning(f"Unable to connect to healthcheck url {self.url}")
        except Exception:
            self.logger.exception("Error while performing healthcheck")
            return False

    def _check_status(self, resp):
        return resp.status_code // 100 == 2 or resp.status_code in self.acceptable_errors


class _NoHealthCheck(HealthCheck):
    def __init__(self, logger):
        super().__init__(None, UrlScheme.GET, logger)

    def is_healthy(self):
        return True



import sys
from datetime import timedelta
import json
import logging
import signal
import time
from kazoo.retry import KazooRetry
from kazoo.client import KazooClient

from serviceprovider.exceptions import StopRangerUpdate
from serviceprovider.health_check import HealthCheck
from serviceprovider.health_check import _NoHealthCheck
from serviceprovider.helper import get_default_logger, default_serialize_func
from serviceprovider.job import Job
from serviceprovider.ranger_models import ClusterDetails, ServiceDetails, HealthcheckStatus, NodeData, ServiceNode

'''
This is a ServiceProvider implementation for doing regular ranger updates on zookeeper
You may run this in background or run it in foreground by blocking your current thread.

Takes care of the following :
- Infinite retry and connection reattempts in case of zk connection issues 
- Proper cleanup of zk connections to get rid of ephemeral nodes
- Proper logging  
= Does continuous health check pings on a particular health check url if required [optional]

'''


def _current_milli_time():
    return round(time.time() * 1000)


def _service_shutdown(signum, frame):
    raise StopRangerUpdate


class _RangerClient(object):
    def __init__(self, zk: KazooClient, cluster_details: ClusterDetails, service_details: ServiceDetails, logger):
        self.zk = zk
        self.cluster_details = cluster_details
        self.service_details = service_details
        self.logger = logger

    def start(self):
        self.zk.start()

    def stop(self):
        self.zk.stop()

    def update_tick(self, status=HealthcheckStatus.HEALTHY):
        node_data = NodeData(self.service_details.environment)
        service_node = ServiceNode(self.service_details.host, self.service_details.port, node_data,
                                   status, _current_milli_time())
        data_bytes = str.encode(json.dumps(service_node.to_dict()))
        self.logger.info(f"Updating with: {str(data_bytes)}")
        if self.zk.exists(self.service_details.get_path()):
            self.zk.set(self.service_details.get_path(), data_bytes)
        else:
            # ensure that you create only ephemeral nodes
            self.zk.ensure_path(self.service_details.get_root_path())
            self.zk.create(self.service_details.get_path(), data_bytes, ephemeral=True)


class RangerServiceProvider(object):
    """
    Initialize this class to be able to start and create a Ranger Updater
    """

    def __init__(self, cluster_details: ClusterDetails, service_details: ServiceDetails,
                 health_check: HealthCheck = None, logger=None):
        """
        :param cluster_details: Zookeeper cluster details
        :param service_details: Service details like name, host, port etc
        :param health_check: health check url details if any
        :param logger: optional logger
        """
        self.cluster_details = cluster_details
        self.service_details = service_details
        self.is_running = False
        self.logger = logger if logger is not None else get_default_logger()
        self.health_check = health_check if health_check is not None else _NoHealthCheck(logger)
        self.ranger_client = _RangerClient(
            KazooClient(hosts=self.cluster_details.zk_string,
                        # proper infinite retries to ensure we handle network flakiness
                        connection_retry=KazooRetry(max_tries=float('inf'), delay=1, max_delay=5)),
            self.cluster_details,
            self.service_details,
            self.logger)
        self.job = None

    def _stop_zk_updates(self):
        if not self.is_running:
            self.logger.info("Already stopped")
            return
        self.logger.info("Stopping all updates to zk and cleaning up..")
        self.ranger_client.stop()
        self.is_running = False

    def _block_main_thread(self):
        signal.signal(signal.SIGTERM, _service_shutdown)
        signal.signal(signal.SIGINT, _service_shutdown)
        while True:
            try:
                time.sleep(1)
            except StopRangerUpdate:
                self._stop_zk_updates()
                break

    def _ranger_update_tick(self):
        """
        Used to perform a single tick update to zookeeper. Handles error scenarios. Does healthcheck if necessary
        """
        try:
            if self.health_check.is_healthy():
                self.ranger_client.update_tick(HealthcheckStatus.HEALTHY)
            else:
                self.ranger_client.update_tick(HealthcheckStatus.UNHEALTHY)
        except Exception:
            self.logger.exception("Error while updating zk")

    def start(self, block=False):
        """
        Creates a Thread that updates zookeeper with service health state updates at regular intervals
        :param block: send block as true if you wish to block the current thread (and wait for an interrupt to stop)
        """
        if self.is_running:
            self.logger.info("Already started")
            return
        self.is_running = True
        self.logger.info(json.dumps(self.cluster_details, default=default_serialize_func))
        self.ranger_client.start()
        self.job = Job(timedelta(seconds=self.cluster_details.update_interval), self._ranger_update_tick)
        self.job.daemon = not block
        self.job.start()
        if block:
            self._block_main_thread()
            self.job.stop()

    def stop(self):
        """
        Stop zookeeper updates
        """
        self.job.stop()
        self.is_running = False



# please install python if it is not present in the system
from setuptools import setup
from serviceprovider import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()


setup(
    name='serviceprovider',
    version=__version__,
    packages=['serviceprovider'],
    license=' License 2.0',
    description='The python equivalent for ranger based service discovery mechanism using zookeeper',
    author='Tushar Naik',
    author_email='tushar.knaik@gmail.com',
    keywords=['ranger', 'zookeeper', 'service discovery', 'periodic task', 'interval', 'periodic job', 'flask style',
              'decorator'],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Tushar-Naik/python-ranger-daemon",
    include_package_data=True,
    py_modules=['serviceprovider'],
    install_requires=[
        'requests',
        'kazoo',
        'python_daemon'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Utilities',
        'Topic :: System :: Networking',
        'Topic :: System :: Distributed Computing'
    ],
)



import logging
import sys


def default_serialize_func(o):
    """
    Use like this: logging.debug(f"print this object: {json.dumps(myobject, indent=4, sort_keys=True, default=default_serialize_func)}")
    """
    if hasattr(o, '__dict__'):
        return o.__dict__
    return f"<could not serialize {o.__class__}>"


def get_default_logger():
    logger = logging.getLogger('ranger')
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(logging.INFO)
    formatter = logging.Formatter('[%(asctime)s] [%(name)s] [%(levelname)s] %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logger.setLevel(logging.INFO)
    return logger



import argparse

from serviceprovider.helper import get_default_logger
from serviceprovider.ranger_models import ClusterDetails, ServiceDetails, UrlScheme
from serviceprovider.service_provider import HealthCheck, RangerServiceProvider

'''
A Python daemon for doing custom Ranger Service Provider registration: 

Writes data in the format (datamodel from ranger):
{"host":"localhost","port":31047,"nodeData":{"environment":"stage"},"healthcheckStatus":"healthy","lastUpdatedTimeStamp":1639044989841}
in path: /namespace/service
at a periodic intervals of --interval (default: 1 second)

How to run this script? 
python3.9 ranger_daemon.py -zk $ZK_CONNECTION_STRING -s $SERVICE_NAME -host $HOST -p $PORT -e $ENV -hcu $HEALTH_CHECK > ranger_daemon.log 

'''


def initial_program_setup(raw_args=None):
    parser = argparse.ArgumentParser(description="Utility to register a service host/port for ")
    parser.add_argument('-zk', '--zkConnectionString', help='zookeeper connection string', required=True)
    parser.add_argument('-n', '--namespace', help='namespace for discovery', default="org")
    parser.add_argument('-s', '--service', help='name of service to be registered', required=True)
    parser.add_argument('-host', '--host', help='hostname of service', required=True)
    parser.add_argument('-p', '--port', help='port of service', required=True, type=int)
    parser.add_argument('-e', '--environment', choices=['stage', 'prod'],
                        help='Environment on which service is running',
                        required=True)
    parser.add_argument('-i', '--interval', help='Update interval in seconds', default=1)
    parser.add_argument('-hcu', '--healthCheckUrl', help='Url where regular health check will be done', default=None)
    parser.add_argument('-hct', '--healthCheckTimeout', help='Url where regular health check will be done', default=0.5)

    print(f'{raw_args}')
    args = parser.parse_args(raw_args)
    logger = get_default_logger()
    return RangerServiceProvider(
        ClusterDetails(args.zkConnectionString, args.interval),
        ServiceDetails(args.host, int(args.port), args.environment, args.namespace, args.service),
        HealthCheck(args.healthCheckUrl, UrlScheme.GET, logger, timeout=args.healthCheckTimeout),
        logger=logger)


def ranger_daemon_trigger(raw_args=None):
    """
    Pass in raw_args if you wana trigger via this method
    eg: ['-zk', 'localhost:2181', '-s', 'myapp', '-host', 'localhost', '-p', '9090', '-e', 'stage', '-hcu', 'http://localhost:9091/healthcheck?pretty=true']
    :param raw_args: sys.argv (arguments to the script)
    """
    ranger_service_provider = initial_program_setup(raw_args)
    ranger_service_provider.start(True)


if __name__ == '__main__':
    ranger_daemon_trigger()



from enum import Enum

"""
includes all models required for the service provider
"""


class HealthcheckStatus(Enum):
    HEALTHY = "healthy"
    UNHEALTHY = "unhealthy"


class UrlScheme(Enum):
    GET = 1
    POST = 2


class NodeData(object):
    def __init__(self, environment):
        self.environment = environment

    def to_dict(self):
        return {"environment": self.environment}


class ServiceNode(object):
    def __init__(self,
                 host,
                 port,
                 node_data: NodeData,
                 healthcheck_status: HealthcheckStatus,
                 last_updated_timestamp):
        self.host = host
        self.port = port
        self.node_data = node_data
        self.last_updated_timestamp = last_updated_timestamp
        self.healthcheck_status = healthcheck_status

    def to_dict(self):
        return {"host": self.host, "port": self.port, "nodeData": self.node_data.to_dict(),
                "healthcheckStatus": self.healthcheck_status.value, "lastUpdatedTimeStamp": self.last_updated_timestamp}


class ServiceDetails(object):
    def __init__(self, host, port, environment, namespace, service_name):
        self.host = host
        self.port = port
        self.environment = environment
        self.namespace = namespace
        self.service_name = service_name

    def get_path(self):
        return f"/{self.namespace}/{self.service_name}/{self.host}:{self.port}"

    def get_root_path(self):
        return f"/{self.namespace}/{self.service_name}"

    def to_dict(self):
        return {"host": self.host, "port": self.port, "environment": self.environment, "namespace": self.namespace,
                "service": self.service_name}


class ClusterDetails(object):
    def __init__(self, zk_string, update_interval):
        self.zk_string = str(zk_string)
        self.update_interval = update_interval

    def to_dict(self):
        return {"zk_string": self.zk_string, "update_interval": self.update_interval}
